﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Text;
using System.Threading.Tasks;
using AvalonRFZero.PRG;
using AvalonRFZero.Libs;
using TorchSharp;
using static TorchSharp.torch;
using System.Diagnostics;

namespace AvalonRFZero.Engines
{
    internal class RFZeroX : IEngine
    {
        public string engineName { set; get; }
        public string engineParams { set; get; }
        public string uniqueID { set; get; } = "Not defined";
        public int RRobinNumber { set; get; }
        public string moveMessage {  set; get; }

        int mctsLoopsMax { set; get; }
        int nodeLoops { set; get; }
        float expParam { set; get; } // Exploration patameter.



        // LibNN4.NetVersion netVersion = LibNN4.NetVersion.Net_02;

        Random random { set; get; }

        TreeNodeMCTS treeRootNode;

        bool NeedLogging { set; get; } = false;
        int NodesSimulayedTotal { set; get; } = 0;
        int ValueSimulationsTotal { set; get; } = 0;
        int PolicySimulationsTotal { set; get; } = 0;
        int TorchThreadsMax { set; get; } = 1; // Set max number of thread for torch Engine, generally 1 is the best value
        long CalculationTime { set; get; } = 0; // Time for calculation in millisexonds.

        enum PosSolved { None, WhiteWins, BlackWins, Draw }

        LibNN8Stars.NetVersion netVersion;

        nn.Module<Tensor, (Tensor, Tensor)> unitedNet;

        bool GPU = false;

        public RFZeroX(string engineName, string engineParams, string NNVersion, int mctsLoopsMax, int nodeLoops,
                string netFilePath, float expParam)
        {
            this.engineName = engineName;
            this.engineParams = engineParams;
            this.mctsLoopsMax = mctsLoopsMax;
            this.nodeLoops = nodeLoops;
            this.expParam = expParam;

            LoadNetsWithParams(NNVersion: NNVersion, NNFilePath: netFilePath);

            random = new Random(StaticMath8Stars.GetSlowRandomSeed());
        }

        public string ShowEngineInfo()
        {
            return "Engine RFZeroX";
        }

        void LoadNetsWithParams(string NNVersion, string NNFilePath)
        {
            // Load net version from string
            this.netVersion = LibNN8Stars.GetUnitedNetEnumFromString(NNVersion);

            this.unitedNet = LibNN8Stars.GetNetStructure(netVersion: netVersion);
            this.unitedNet.load(NNFilePath);

            if (this.GPU == true)
                unitedNet = unitedNet.cuda();

            // Switch to evaluation mode.
            unitedNet.eval();

            // Set number of Threads for torch equal to 1.
            if (this.GPU == false)
                torch.set_num_threads(6);
        }

        public int[] ShowBestMove(byte[] Position)
        {
            if (StaticMath8Stars.FiguresOnTheBoard(Position) == 4) // Start position, need to make random move here..
            {
                int[] rndMove = StaticMath8Stars.TakeRandomMove(rnd: random, Position: Position);
                return rndMove;
            }

            (int moveRow, int moveCol, List<int[]> MctsData) = MainMCTSLoop(Position);

            return new int[] { moveRow, moveCol };
        }

        public BestMoveData ShowsBestMoveAndStat(byte[] Position)
        {
            float NodesSec = ((float)this.NodesSimulayedTotal / (float)this.CalculationTime) * 1000;

            // double timeToThink = Math.Round((this.CalculationTime) / 1000),2);
            double timeToThink = Math.Round(((double)this.CalculationTime / 1000), 2);

            if (StaticMath8Stars.FiguresOnTheBoard(Position) == 4) // Start position, need to make random move here..
            {
                int[] rndMove = StaticMath8Stars.TakeRandomMove(rnd: random, Position: Position);

                BestMoveData bestMoveData = new BestMoveData()
                {
                    BestMove = rndMove,
                    MoveInfo = $"MCTS loops: {this.NodesSimulayedTotal} Sec: {timeToThink} ",
                };

                return bestMoveData;
            }

            else
            {
                (int moveRow, int moveCol, List<int[]> MctsData) = MainMCTSLoop(Position);

                BestMoveData bestMoveData = new BestMoveData()
                {
                    BestMove = new int[] { moveRow, moveCol },
                    MoveInfo = $"Nodes total: {this.NodesSimulayedTotal} Sec: {timeToThink} ",
                };

                return bestMoveData;
            }
        }

        (int moveRow, int moveCol, List<int[]> MctsData) MainMCTSLoop(byte[] Position)
        {
            Stopwatch sw = Stopwatch.StartNew();

            this.NodesSimulayedTotal = 0;
            this.ValueSimulationsTotal = 0;
            this.PolicySimulationsTotal = 0;

            byte[] RootPosition = StaticMath8Stars.CopyPosition(Position);

            TreeNodeMCTS rootNode = new TreeNodeMCTS(null, RootPosition, 0);

            for (int loopIndex = 0; loopIndex < mctsLoopsMax; loopIndex++)
            {
                TreeNodeMCTS SelectedNode = null;

                SelectNode(rootNode, out SelectedNode);

                ExpandNode(SelectedNode);

                SimulateNode(SelectedNode);

                BackPropagate(SelectedNode);
            }

            sw.Stop();
            CalculationTime = sw.ElapsedMilliseconds;

            // Log some info here.....
            treeRootNode = rootNode;

            // ***** No logging for now....
            // AddNodesInfoToFile();

            int MostSimulatedNodeIndex = FindMaxSimulationNodeIndex(rootNode);
            TreeNodeMCTS bestNode = rootNode.ChildNodesList[MostSimulatedNodeIndex];
            List<int[]> MctsData = FindMctsValues(rootNode);

            // MstsVal = new MctsValue(WhiteWins: bestNode.whiteWins, Blackwins: bestNode.blackWins, DrawsCount: bestNode.draws);

            return (moveRow: bestNode.lastMoveRow, moveCol: bestNode.lastMoveCol, MctsData: MctsData);

        }



        void SelectNode(TreeNodeMCTS Node, out TreeNodeMCTS selectedNode)
        {
            // We go to nextNode if some child nodes exist AND all of them have been simulated and backpropagated.
            if (Node.ChildNodesList.Count > 0 && Node.ChildNodesList.Count == Node.SimulatedChildNodesCount)
            {
                int bestNodeIndex = FindNextNodeIndex(Node);
                TreeNodeMCTS nextNode = Node.ChildNodesList[bestNodeIndex];
                SelectNode(nextNode, out selectedNode);
            }
            else
            {
                selectedNode = Node;
            }
        }

        void ExpandNode(TreeNodeMCTS Node)
        {
            if (Node.ChildNodesList.Count > 0) // Node already has all child nodes list.
                return;

            if (Node.nodePosition[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.GameOver)
                return; // Game over here, expansion is impossible

            // int[,] allMovesRowCol = StaticMath8Stars.GetAllMoves2D(Node.nodePosition);

            List<int[]> allMovesRowCol = StaticMath8Stars.GetAllMoves(Node.nodePosition);

            for (int moveIndx = 0; moveIndx < allMovesRowCol.Count; moveIndx++)
            {
                int moveRow = allMovesRowCol[moveIndx][0];
                int moveCol = allMovesRowCol[moveIndx][1];

                byte[] newPosition = StaticMath8Stars.GetNextPosition(Node.nodePosition, moveRow, moveCol);

                TreeNodeMCTS nextNode = new TreeNodeMCTS(Node, newPosition, Node.nodeDepth + 1);
                nextNode.lastMoveRow = moveRow;
                nextNode.lastMoveCol = moveCol;

                Node.ChildNodesList.Add(nextNode);
            }
        }

        void SimulateNode(TreeNodeMCTS Node)
        {
            if (Node.ChildNodesList.Count == 0) // Simulation happens after expansion, if no child nodes - game over!
                return;

            if (Node.BackPropOrder.Count > 0) // Array BackPropOrder already ecists, all the simulation data set, no need to simulate again
                return;

            // Creating List of positions            
            List<byte[]> posList = Node.ExtractChildPositions();

            // Add root position to the List
            posList.Add(Node.nodePosition);

            // Find all position moves

            List<List<int[]>> allMovesList = new List<List<int[]>>(); 

            for (int i = 0; i < posList.Count; i++) 
            {
                // int[,] posMovesRowCol = StaticMath03.GetAllMoves2D(posList[i]); // This array can be NULL! Position have vo valid moves! Game over!

                List<int[]> posMovesRowCol = StaticMath8Stars.GetAllMoves(posList[i]);

                if (posMovesRowCol.Count == 0) 
                {
                    // Make some random array here, positions with no moves (game over) will ve filtered
                    // inside function SimulatePositionValues 
                    // Pos array can not contain null move values.
                    
                    posMovesRowCol = new List<int[]> { new int[2] {0, 0} }; 
                }

                allMovesList.Add(posMovesRowCol);
            }

            // using Tensor xTensorNet = LibNN4.GetTensorSuper(ListOfPositions: posList, allMovesPositions: allMovesList, GPU: this.GPU);

            using Tensor xTensorNet = LibNN8Stars.CreateNetTensor(netVersion, posList, allMovesList, GPU: this.GPU);

            (Tensor yTensorValNet, Tensor yTensorPolNet) = unitedNet.forward(xTensorNet);

            yTensorValNet = yTensorValNet.cpu();
            yTensorPolNet = yTensorPolNet.cpu();

            SimulatePositionValues(rnd: this.random, childNodes: Node.ChildNodesList, outputValueNet: yTensorValNet, 
                NSimulations: this.nodeLoops);

            CalculatePolicyNetValues(Node: Node, outputPolicyNet: yTensorPolNet);

            // Ater Monte Carlo simulation we need to vreate BackPropagate array for good backpropagation order.
            CreateBackPropOrderArray(Node);

            yTensorValNet.Dispose();
            yTensorPolNet.Dispose();
        }

        void BackPropagate(TreeNodeMCTS Node)
        {
            if (Node.ChildNodesList.Count > 0)
            {
                int BackPropagateIndex = Node.BackPropOrder[Node.SimulatedChildNodesCount];

                BackPropagateNode(Node.ChildNodesList[BackPropagateIndex],
                    Node.ChildNodesList[BackPropagateIndex].whiteWinsDirect,
                    Node.ChildNodesList[BackPropagateIndex].blackWinsDirect,
                    Node.ChildNodesList[BackPropagateIndex].drawsDirect);

                Node.SimulatedChildNodesCount++;
            }

            else // We are in a Final Node now, game over Here... we need make backpropagation anyway.
            {
                (int WhiteWins, int BlackWins, int Draws) = ResultsOfFinalNode(Node);

                BackPropagateNode(Node, WhiteWins, BlackWins, Draws);
            }

            this.NodesSimulayedTotal++;
        }

        void BackPropagateNode(TreeNodeMCTS Node, int WhiteWins, int BlackWins, int Draws)
        {
            Node.whiteWins += WhiteWins;
            Node.blackWins += BlackWins;
            Node.draws += Draws;

            Node.nVisits++;

            if (Node.rootNode != null)
            {
                BackPropagateNode(Node.rootNode, WhiteWins, BlackWins, Draws);
            }
        }

        void SimulatePositionValues(Random rnd, List<TreeNodeMCTS> childNodes, Tensor outputValueNet, int NSimulations)
        {
            for (int pIndex = 0; pIndex < childNodes.Count; pIndex++)
            {
                byte[] Position = childNodes[pIndex].nodePosition;

                if (Position[StaticMath8Stars.MoveTurnIndex] != StaticMath8Stars.GameOver) // Simulate positions here...
                {
                    float probWinNominalWhite = outputValueNet[pIndex][0].item<float>();
                    float probWinNominalBlack = outputValueNet[pIndex][1].item<float>();

                    float WhiteLen = probWinNominalWhite;
                    float BlackLen = probWinNominalWhite + probWinNominalBlack;

                    int NomWhiteWin = 0;
                    int NomBlackWin = 0;
                    int NomDraws = 0;

                    for (int i = 0; i < NSimulations; i++)
                    {
                        Double rndValue = rnd.NextDouble();

                        if (rndValue <= WhiteLen)
                            NomWhiteWin += 1;
                        else if (rndValue > WhiteLen && rndValue <= BlackLen)
                            NomBlackWin += 1;
                        else
                            NomDraws++;
                    }

                    if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.WhiteTurn)
                    {
                        childNodes[pIndex].whiteWinsDirect = NomWhiteWin;
                        childNodes[pIndex].blackWinsDirect = NomBlackWin;
                        childNodes[pIndex].drawsDirect = NomDraws;
                    }

                    else
                    {
                        childNodes[pIndex].whiteWinsDirect = NomBlackWin;
                        childNodes[pIndex].blackWinsDirect = NomWhiteWin;
                        childNodes[pIndex].drawsDirect = NomDraws;
                    }
                }

                else // No need to simulate if game is over. Set results directly.
                {
                    (int whiteCount, int blackCount) = StaticMath8Stars.StaticResult(Position);

                    if (whiteCount > blackCount) // White wins here...
                    {
                        childNodes[pIndex].whiteWinsDirect = NSimulations;
                        childNodes[pIndex].blackWinsDirect = 0;
                        childNodes[pIndex].drawsDirect = 0;
                    }

                    else if (whiteCount < blackCount) // Black wins here...
                    {
                        childNodes[pIndex].whiteWinsDirect = 0;
                        childNodes[pIndex].blackWinsDirect = NSimulations;
                        childNodes[pIndex].drawsDirect = 0;
                    }

                    else // Draw result here
                    {
                        childNodes[pIndex].whiteWinsDirect = 0;
                        childNodes[pIndex].blackWinsDirect = 0;
                        childNodes[pIndex].drawsDirect = NSimulations;
                    }
                }
            }
        }


        void CalculatePolicyNetValues(TreeNodeMCTS Node, Tensor outputPolicyNet) 
        {
            long nodeTensorIndex = outputPolicyNet.shape[0] - 1;

            for (int moveIndx = 0; moveIndx < Node.ChildNodesList.Count; moveIndx++)
            {
                TreeNodeMCTS childNode = Node.ChildNodesList[moveIndx];

                int moveRow = childNode.lastMoveRow;
                int moveCol = childNode.lastMoveCol;

                childNode.PriorProb = outputPolicyNet[nodeTensorIndex][moveRow * StaticMath8Stars.BoardSize + moveCol].item<float>();
            }
        }

        int FindNextNodeIndex(TreeNodeMCTS Node) 
        {
            float best_UCB_Value = -1f;
            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                UCB_NodeValue(Node.ChildNodesList[i]);
            }

            // Now find maximum usbValue
            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                if (Node.ChildNodesList[i].ucbValue > best_UCB_Value)
                {
                    best_UCB_Value = Node.ChildNodesList[i].ucbValue;
                }
            }

            // Now find all the nodes with best UCB Value, can be more than ONE.
            List<int> bestNodeIndexes = new List<int>();
            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                if (Node.ChildNodesList[i].ucbValue == best_UCB_Value)
                {
                    bestNodeIndexes.Add(i);
                }
            }

            // Finally, if only ONE best Node found, return its index,
            // otherwise find random lidex from the bestNodeIndexes

            if (bestNodeIndexes.Count > 1)
            {
                return bestNodeIndexes[0];
            }
            
            else
            {
                return bestNodeIndexes[random.Next(bestNodeIndexes.Count)];
            }
        }

        (int WhiteWins, int BlackWins, int Draws) ResultsOfFinalNode(TreeNodeMCTS treeNodeMCST)
        {
            // We need to check, if position is GAME OVER for now,
            // and only after that analize it with NN module.
            int simulationsTotal = this.nodeLoops;

            // ArraySegment<int> CurrentPosition = treeNodeMCST.nodePosition;

            (int whiteCount, int blackCount) = StaticMath8Stars.StaticResult(treeNodeMCST.nodePosition);

            if (whiteCount > blackCount)
                return (WhiteWins: simulationsTotal, BlackWins: 0, Draws: 0);
            else if (whiteCount < blackCount)
                return (WhiteWins: 0, BlackWins: simulationsTotal, Draws: 0);
            else
                return (WhiteWins: 0, BlackWins: 0, Draws: simulationsTotal);
        }

        void CreateBackPropOrderArray(TreeNodeMCTS Node) 
        {
            DataTable dataTable = new DataTable();

            dataTable.Columns.Add("NodeIndex", typeof(int));
            dataTable.Columns.Add("NodeOrderValue", typeof(float));

            float PriorProbMulKoeff = 1f;

            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                DataRow newRow = dataTable.Rows.Add();
                newRow["NodeIndex"] = i;

                TreeNodeMCTS childNode = Node.ChildNodesList[i];

                int SimsTotal = childNode.whiteWinsDirect + childNode.blackWinsDirect + childNode.drawsDirect;
                float SimWinRes = 0;

                if (Node.NodeMoveTurn() == StaticMath8Stars.WhiteTurn)
                    SimWinRes = (childNode.whiteWinsDirect + 0.5f * childNode.drawsDirect) / SimsTotal;
                else
                    SimWinRes = (childNode.blackWinsDirect + 0.5f * childNode.drawsDirect) / SimsTotal;

                newRow["NodeOrderValue"] = SimWinRes + PriorProbMulKoeff * Node.ChildNodesList[i].PriorProb;
            }

            // Sort data here...
            DataView view = new DataView(dataTable);
            view.Sort = "NodeOrderValue DESC";

            DataTable sortedTable = view.ToTable();

            for (int i = 0; i < sortedTable.Rows.Count; i++)
            {
                Node.BackPropOrder.Add((int)(sortedTable.Rows[i]["NodeIndex"]));
            }
        }

        void UCB_NodeValue(TreeNodeMCTS Node) 
        {
            // float ExpParam = 1; // UCB Exploration parameter.
            float w_i;
            int n_i;

            if (Node.rootNode.nodePosition[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.WhiteTurn)
                w_i = (float)Node.whiteWins + 0.5f * Node.draws;
            else
                w_i = (float)Node.blackWins + 0.5f * Node.draws;

            n_i = Node.whiteWins + Node.blackWins + Node.draws;

            float Q = w_i / (float)n_i;
            float U = this.expParam * Node.PriorProb * (float)(Math.Sqrt(Node.rootNode.nVisits) / (1 + Node.nVisits));

            Node.ucbValue = Q + U;
        }

        List<int[]> FindMctsValues(TreeNodeMCTS Node)
        {
            List<int[]> MctsData = new List<int[]>();

            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                TreeNodeMCTS childNode = Node.ChildNodesList[i];

                int[] childNodeValues = new int[] { childNode.lastMoveRow, childNode.lastMoveCol, childNode.whiteWins + childNode.blackWins + childNode.draws };
                MctsData.Add(childNodeValues);
            }

            return MctsData;
        }

        int FindMaxSimulationNodeIndex(TreeNodeMCTS Node)
        {
            // Now find maximum number of nodes simulation.
            int simulationsMax = 0;

            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                int ChildNodeSimulatios = NodeSimulatios(Node.ChildNodesList[i]);

                if (ChildNodeSimulatios > simulationsMax)
                {
                    simulationsMax = ChildNodeSimulatios;
                }
            }

            // Now find all the nodes with max simulations, can be more than ONE.
            List<int> bestNodeIndexes = new List<int>();
            for (int i = 0; i < Node.ChildNodesList.Count; i++)
            {
                if (NodeSimulatios(Node.ChildNodesList[i]) == simulationsMax)
                {
                    bestNodeIndexes.Add(i);
                }
            }

            // Finally, if only ONE best Node found, return its index,
            // otherwise find random lidex from the bestNodeIndexes

            if (bestNodeIndexes.Count > 1)
            {
                return bestNodeIndexes[0];
            }
            else
            {
                return bestNodeIndexes[random.Next(bestNodeIndexes.Count)];
            }

            int NodeSimulatios(TreeNodeMCTS Node)
            {
                return Node.whiteWins + Node.blackWins + Node.draws;
            }
        }

        class TreeNodeMCTS
        {
            public TreeNodeMCTS rootNode;

            public int nodeDepth;
            public byte[] nodePosition;
            public List<TreeNodeMCTS> ChildNodesList = new List<TreeNodeMCTS>();

            // public LibNN2.MoveProb[] movesProb = null;
            public List<int> BackPropOrder = new List<int>();
            // public List<SimReasult> SimChildResultsAll = new List<SimReasult>();

            public int whiteWins = 0;
            public int blackWins = 0;
            public int draws = 0;

            public int whiteWinsDirect; // Direct evaluation with NN or by the game rules.
            public int blackWinsDirect; // Direct evaluation with NN or by the game rules.
            public int drawsDirect; // Direct evaluation with NN or by the game rules.

            public int nVisits = 0;

            public float ucbValue = 0;

            public float PriorProb = 0f;

            public int SimulatedChildNodesCount = 0;

            public int lastMoveRow = -1;
            public int lastMoveCol = -1;

            public PosSolved posSolved = PosSolved.None;

            public TreeNodeMCTS(TreeNodeMCTS rootNode, byte[] nodePosition, int nodeDepth)
            {
                this.rootNode = rootNode;
                this.nodeDepth = nodeDepth;
                this.nodePosition = nodePosition;
            }

            public int NodeMoveTurn()
            {
                return nodePosition[StaticMath8Stars.MoveTurnIndex];
            }

            public List<byte[]> ExtractChildPositions() 
            {
                // Creating List of positions            
                List<byte[]> posList = new List<byte[]>(this.ChildNodesList.Count);

                for (int i = 0; i < this.ChildNodesList.Count; i++)
                {
                    posList.Add(this.ChildNodesList[i].nodePosition);
                }

                return posList;
            }

            public void SetDirectResults(int WhiteWins, int BlackWins, int DrawsCount, bool GameOver)
            {
                this.whiteWinsDirect = WhiteWins;
                this.blackWinsDirect = BlackWins;
                this.drawsDirect = DrawsCount;

                if (GameOver == true && (WhiteWins > BlackWins))
                    this.posSolved = PosSolved.WhiteWins;

                if (GameOver == true && (BlackWins > WhiteWins))
                    this.posSolved = PosSolved.BlackWins;

                if (GameOver == true && WhiteWins == BlackWins)
                    this.posSolved = PosSolved.Draw;
            }
        }

        class SimReasult
        {
            public bool GameOver = false;
            public int WhiteWins = 0;
            public int BlackWins = 0;
            public int DrawsCount = 0;

            public void SetResults(int WhiteWins, int BlackWins, int DrawsCount)
            {
                this.WhiteWins = WhiteWins;
                this.BlackWins = BlackWins;
                this.DrawsCount = DrawsCount;
            }
        }
    }
}
