﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using TorchSharp;
using static TorchSharp.torch;

namespace AvalonRFZero.Libs
{
    class LibNN8Stars
    {
        internal enum NetVersion { Net_01, Net_02, Net_03, Net_04, Net_05 }

        internal static nn.Module<Tensor, (Tensor, Tensor)> GetNetStructure(NetVersion netVersion)
        {
            switch (netVersion)
            {
                case NetVersion.Net_01: return new ComplexNet01();
                case NetVersion.Net_02: return new ComplexNet02();
                case NetVersion.Net_03: return new ComplexNet03();
                case NetVersion.Net_04: return new ComplexNet04();
                case NetVersion.Net_05: return new ComplexNet05();


                default: return null;
            }
        }

        internal static Tensor CreateNetTensor(NetVersion netVersion, List<byte[]> ListOfPositions, List<List<int[]>> allMoves, bool GPU)
        {
            switch (netVersion)
            {
                case NetVersion.Net_01: return GetTensor01(ListOfPositions, allMoves, GPU);
                case NetVersion.Net_02: return GetTensor01(ListOfPositions, allMoves, GPU);
                case NetVersion.Net_03: return GetTensor01(ListOfPositions, allMoves, GPU);
                case NetVersion.Net_04: return GetTensor01(ListOfPositions, allMoves, GPU);
                case NetVersion.Net_05: return GetTensor01(ListOfPositions, allMoves, GPU);

                default: return null;
            }
        }

        // ************************************  Available Tensors.
        private static Tensor GetTensor01(List<byte[]> ListOfPositions, List<List<int[]>> allMovesPositions, bool GPU)
        {
            int objStart = 0;
            int objEnd = ListOfPositions.Count - 1;
            int posTotal = ListOfPositions.Count; // Total positions for current batch.

            int boardSize = StaticMath8Stars.BoardSize;

            float[,,,] cnnBatchArray = new float[posTotal, 3, boardSize, boardSize]; // Tensor element contains 3 planes.

            for (int objIndx = objStart; objIndx <= objEnd; objIndx++)
            {
                // **** Nominal white means side with current move turn
                int NominalWhite = ListOfPositions[objIndx][StaticMath8Stars.MoveTurnIndex];

                for (int row = 0; row < boardSize; row++)
                {
                    for (int col = 0; col < boardSize; col++)
                    {
                        int pIndx = row * boardSize + col;
                        int posValue = ListOfPositions[objIndx][pIndx];

                        if (posValue == StaticMath8Stars.EmptyFig)
                            continue;

                        else if (posValue == NominalWhite)
                        {
                            cnnBatchArray[objIndx, 0, row, col] = 1; // Figures with move turn keep in the First 2D Plane.
                        }

                        else
                        {
                            cnnBatchArray[objIndx, 1, row, col] = 1; // Figures with Opposite move turn keep in the Second 2D Plane.
                        }
                    }
                }

                List<int[]> posMoves = allMovesPositions[objIndx];

                for (int moveIndx = 0; moveIndx < posMoves.Count; moveIndx++)
                {
                    int moveRow = posMoves[moveIndx][0];
                    int moveCol = posMoves[moveIndx][1];

                    cnnBatchArray[objIndx, 2, moveRow, moveCol] = 1; // For the fird plane we set 1 to the point with available move and 0 to the point with no move
                }
            }

            if (GPU == true)
                return torch.tensor(cnnBatchArray, device: CUDA);
            else
                return torch.tensor(cnnBatchArray);
        }

        private static Tensor GetTensorBits(List<byte[]> ListOfPositions, List<List<int[]>> allMovesPositions, bool GPU)
        {
            int objStart = 0;
            int objEnd = ListOfPositions.Count - 1;
            int posTotal = ListOfPositions.Count; // Total positions for current batch.

            int boardSize = StaticMath8Stars.BoardSize;

            bool[,,,] cnnBatchArray = new bool[posTotal, 3, boardSize, boardSize]; // Tensor element contains 3 planes.

            for (int objIndx = objStart; objIndx <= objEnd; objIndx++)
            {
                // **** Nominal white means side with current move turn
                int NominalWhite = ListOfPositions[objIndx][StaticMath8Stars.MoveTurnIndex];

                for (int row = 0; row < boardSize; row++)
                {
                    for (int col = 0; col < boardSize; col++)
                    {
                        int pIndx = row * boardSize + col;
                        int posValue = ListOfPositions[objIndx][pIndx];

                        if (posValue == StaticMath8Stars.EmptyFig)
                            continue;

                        else if (posValue == NominalWhite)
                        {
                            cnnBatchArray[objIndx, 0, row, col] = true; // Figures with move turn keep in the First 2D Plane.
                        }

                        else
                        {
                            cnnBatchArray[objIndx, 1, row, col] = true; // Figures with Opposite move turn keep in the Second 2D Plane.
                        }
                    }
                }

                List<int[]> posMoves = allMovesPositions[objIndx];

                for (int moveIndx = 0; moveIndx < posMoves.Count; moveIndx++)
                {
                    int moveRow = posMoves[moveIndx][0];
                    int moveCol = posMoves[moveIndx][1];

                    cnnBatchArray[objIndx, 2, moveRow, moveCol] = true; // For the fird plane we set 1 to the point with available move and 0 to the point with no move
                }
            }

            if (GPU == true)
                return torch.tensor(cnnBatchArray, dtype: ScalarType.Bool, device: CUDA);
            else
                return torch.tensor(cnnBatchArray, dtype: ScalarType.Bool);
        }

        // ************************************  Available NETs

        public class ComplexNet01 : nn.Module<Tensor, (Tensor, Tensor)> // Based on ComplexNet10
        {
            public ComplexNet01()
            : base(nameof(ComplexNet01))
            {
                List<(string, nn.Module<Tensor, Tensor>)> modulesValNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesValNet.Add(("Flat", nn.Flatten()));
                modulesValNet.Add(("LinVal01", linVal_1));
                modulesValNet.Add(("Relu01", nn.ReLU()));
                modulesValNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesValNet.Add(("LinVal02", linVal_2));
                modulesValNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersValNet = nn.Sequential(modulesValNet);

                List<(string, nn.Module<Tensor, Tensor>)> modulesPolNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesPolNet.Add(("Flat", nn.Flatten()));
                modulesPolNet.Add(("LinPol01", linPol_1));
                modulesPolNet.Add(("Relu01", nn.ReLU()));
                modulesPolNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesPolNet.Add(("LinPol02", linPol_2));
                modulesPolNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersPolNet = nn.Sequential(modulesPolNet);

                RegisterComponents();
            }

            // Alpha - Zero style NET here...
            public override (Tensor, Tensor) forward(Tensor input)
            {
                // Start common block
                using var cnnStart = conv_start.forward(input);
                using var relStart = nn.functional.relu(cnnStart);

                using var outRes01 = resNetBlock.forward(relStart);
                using var outRes02 = resNetBlock.forward(outRes01);
                using var outRes03 = resNetBlock.forward(outRes02);


                // Value Net block

                using var outConvEndValNet = conv_endValNet.forward(outRes03);

                // Policy Net block

                using var outConvEndPolNet = conv_endPolNet.forward(outRes03);

                return (LinLayersValNet.forward(outConvEndValNet), LinLayersPolNet.forward(outConvEndPolNet));
            }


            // **************  ResMet Unit.
            class ResNetUnit : nn.Module<Tensor, Tensor>
            {
                public ResNetUnit()
                    : base(nameof(ResNetUnit))
                {
                    RegisterComponents();
                }

                public override Tensor forward(Tensor input)
                {
                    using var resnet = input;

                    using var outputConv01 = ConvModule.forward(input);
                    using var outputRelu01 = nn.functional.relu(outputConv01);

                    using var outputConv02 = ConvModule.forward(outputRelu01);

                    return (nn.functional.relu(outputConv02) + resnet);
                }

                private TorchSharp.Modules.Conv2d ConvModule = nn.Conv2d(dtype: ScalarType.Float32,
                    in_channels: cnnFiltersResBlock, out_channels: cnnFiltersResBlock, kernel_size: 3, padding: 1,
                    padding_mode: PaddingModes.Zeros, stride: 1);
            }

            // Board size
            static int BoardSize = 10;
            static int BoardPoints = BoardSize * BoardSize;

            // Res net params.
            static int cnnFiltersResBlock = 256;

            // End of convolutiobal block.
            static int convEndFiltersValNet = 3;
            static int convEndFiltersPolNet = 3;

            // Variables value block
            static int FlattenOutValNet = BoardPoints * convEndFiltersValNet;
            static int FlattenOutPolNet = BoardPoints * convEndFiltersPolNet;

            static float DropOut = 0.1f;

            static int inputChannels = 3;

            static int LayerValNeurons = 256; // 256 by default
            static int OutputValValues = 3;

            static int LayerPolNeurons = 256; // 256 by default
            static int OutputPolValues = BoardPoints;

            private readonly nn.Module<Tensor, Tensor> LinLayersValNet;
            private readonly nn.Module<Tensor, Tensor> LinLayersPolNet;


            private nn.Module<Tensor, Tensor> conv_start = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: inputChannels, out_channels: cnnFiltersResBlock, kernel_size: 3,
                padding: 1, padding_mode: PaddingModes.Zeros, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endValNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersValNet, kernel_size: 1, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endPolNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersPolNet, kernel_size: 1, stride: 1);


            private nn.Module<Tensor, Tensor> resNetBlock = new ResNetUnit();

            private nn.Module<Tensor, Tensor> flat_01 = nn.Flatten();

            private nn.Module<Tensor, Tensor> linVal_1 = nn.Linear(FlattenOutValNet, LayerValNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linVal_2 = nn.Linear(LayerValNeurons, OutputValValues, dtype: ScalarType.Float32);

            private nn.Module<Tensor, Tensor> linPol_1 = nn.Linear(FlattenOutPolNet, LayerPolNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linPol_2 = nn.Linear(LayerPolNeurons, OutputPolValues, dtype: ScalarType.Float32);
        }

        public class ComplexNet02 : nn.Module<Tensor, (Tensor, Tensor)> // Based on ComplexNet10
        {
            public ComplexNet02()
            : base(nameof(ComplexNet02))
            {
                List<(string, nn.Module<Tensor, Tensor>)> modulesValNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesValNet.Add(("Flat", nn.Flatten()));
                modulesValNet.Add(("LinVal01", linVal_1));
                modulesValNet.Add(("Relu01", nn.ReLU()));
                modulesValNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesValNet.Add(("LinVal02", linVal_2));
                modulesValNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersValNet = nn.Sequential(modulesValNet);

                List<(string, nn.Module<Tensor, Tensor>)> modulesPolNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesPolNet.Add(("Flat", nn.Flatten()));
                modulesPolNet.Add(("LinPol01", linPol_1));
                modulesPolNet.Add(("Relu01", nn.ReLU()));
                modulesPolNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesPolNet.Add(("LinPol02", linPol_2));
                modulesPolNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersPolNet = nn.Sequential(modulesPolNet);

                RegisterComponents();
            }

            // Alpha - Zero style NET here...
            public override (Tensor, Tensor) forward(Tensor input)
            {
                // Start common block
                using var cnnStart = conv_start.forward(input);
                using var relStart = nn.functional.relu(cnnStart);

                using var outRes01 = resNetBlock.forward(relStart);
                using var outRes02 = resNetBlock.forward(outRes01);
                using var outRes03 = resNetBlock.forward(outRes02);
                using var outRes04 = resNetBlock.forward(outRes03);
                using var outRes05 = resNetBlock.forward(outRes04);
                using var outRes06 = resNetBlock.forward(outRes05);
                using var outRes07 = resNetBlock.forward(outRes06);



                // Value Net block

                using var outConvEndValNet = conv_endValNet.forward(outRes07);

                // Policy Net block

                using var outConvEndPolNet = conv_endPolNet.forward(outRes07);

                return (LinLayersValNet.forward(outConvEndValNet), LinLayersPolNet.forward(outConvEndPolNet));
            }


            // **************  ResMet Unit.
            class ResNetUnit : nn.Module<Tensor, Tensor>
            {
                public ResNetUnit()
                    : base(nameof(ResNetUnit))
                {
                    RegisterComponents();
                }

                public override Tensor forward(Tensor input)
                {
                    using var resnet = input;

                    using var outputConv01 = ConvModule.forward(input);
                    using var outputRelu01 = nn.functional.relu(outputConv01);

                    using var outputConv02 = ConvModule.forward(outputRelu01);

                    return (nn.functional.relu(outputConv02) + resnet);
                }

                private TorchSharp.Modules.Conv2d ConvModule = nn.Conv2d(dtype: ScalarType.Float32,
                    in_channels: cnnFiltersResBlock, out_channels: cnnFiltersResBlock, kernel_size: 3, padding: 1,
                    padding_mode: PaddingModes.Zeros, stride: 1);
            }

            // Board size
            static int BoardSize = 10;
            static int BoardPoints = BoardSize * BoardSize;

            // Res net params.
            static int cnnFiltersResBlock = 256;

            // End of convolutiobal block.
            static int convEndFiltersValNet = 2; // Reduce end conv filters to 2 (original has 3)
            static int convEndFiltersPolNet = 2; // Reduce end conv filters to 2 (original has 3)

            // Variables value block
            static int FlattenOutValNet = BoardPoints * convEndFiltersValNet;
            static int FlattenOutPolNet = BoardPoints * convEndFiltersPolNet;

            static float DropOut = 0.1f;

            static int inputChannels = 3;

            static int LayerValNeurons = 256; // 256 by default
            static int OutputValValues = 3;

            static int LayerPolNeurons = 256; // 256 by default
            static int OutputPolValues = BoardPoints;

            private readonly nn.Module<Tensor, Tensor> LinLayersValNet;
            private readonly nn.Module<Tensor, Tensor> LinLayersPolNet;


            private nn.Module<Tensor, Tensor> conv_start = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: inputChannels, out_channels: cnnFiltersResBlock, kernel_size: 3,
                padding: 1, padding_mode: PaddingModes.Zeros, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endValNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersValNet, kernel_size: 1, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endPolNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersPolNet, kernel_size: 1, stride: 1);


            private nn.Module<Tensor, Tensor> resNetBlock = new ResNetUnit();

            private nn.Module<Tensor, Tensor> flat_01 = nn.Flatten();

            private nn.Module<Tensor, Tensor> linVal_1 = nn.Linear(FlattenOutValNet, LayerValNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linVal_2 = nn.Linear(LayerValNeurons, OutputValValues, dtype: ScalarType.Float32);

            private nn.Module<Tensor, Tensor> linPol_1 = nn.Linear(FlattenOutPolNet, LayerPolNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linPol_2 = nn.Linear(LayerPolNeurons, OutputPolValues, dtype: ScalarType.Float32);
        }


        public class ComplexNet03 : nn.Module<Tensor, (Tensor, Tensor)> // Based on ComplexNet10
        {
            public ComplexNet03()
            : base(nameof(ComplexNet03))
            {
                List<(string, nn.Module<Tensor, Tensor>)> modulesValNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesValNet.Add(("Flat", nn.Flatten()));
                modulesValNet.Add(("LinVal01", linVal_1));
                modulesValNet.Add(("Relu01", nn.ReLU()));
                modulesValNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesValNet.Add(("LinVal02", linVal_2));
                modulesValNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersValNet = nn.Sequential(modulesValNet);

                List<(string, nn.Module<Tensor, Tensor>)> modulesPolNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesPolNet.Add(("Flat", nn.Flatten()));
                modulesPolNet.Add(("LinPol01", linPol_1));
                modulesPolNet.Add(("Relu01", nn.ReLU()));
                modulesPolNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesPolNet.Add(("LinPol02", linPol_2));
                modulesPolNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersPolNet = nn.Sequential(modulesPolNet);

                RegisterComponents();
            }

            // Alpha - Zero style NET here...
            public override (Tensor, Tensor) forward(Tensor input)
            {
                // Start common block
                using var cnnStart = conv_start.forward(input);
                using var relStart = nn.functional.relu(cnnStart);

                using var outRes01 = resNetBlock.forward(relStart);

                // Value Net block

                using var outConvEndValNet = conv_endValNet.forward(outRes01);

                // Policy Net block

                using var outConvEndPolNet = conv_endPolNet.forward(outRes01);

                return (LinLayersValNet.forward(outConvEndValNet), LinLayersPolNet.forward(outConvEndPolNet));
            }


            // **************  ResMet Unit.
            class ResNetUnit : nn.Module<Tensor, Tensor>
            {
                public ResNetUnit()
                    : base(nameof(ResNetUnit))
                {
                    RegisterComponents();
                }

                public override Tensor forward(Tensor input)
                {
                    using var resnet = input;

                    using var outputConv01 = ConvModule.forward(input);
                    using var outputRelu01 = nn.functional.relu(outputConv01);

                    using var outputConv02 = ConvModule.forward(outputRelu01);

                    return (nn.functional.relu(outputConv02) + resnet);
                }

                private TorchSharp.Modules.Conv2d ConvModule = nn.Conv2d(dtype: ScalarType.Float32,
                    in_channels: cnnFiltersResBlock, out_channels: cnnFiltersResBlock, kernel_size: 3, padding: 1,
                    padding_mode: PaddingModes.Zeros, stride: 1);
            }

            // Board size
            static int BoardSize = 10;
            static int BoardPoints = BoardSize * BoardSize;

            // Res net params.
            static int cnnFiltersResBlock = 256;

            // End of convolutiobal block.
            static int convEndFiltersValNet = 3;
            static int convEndFiltersPolNet = 3;

            // Variables value block
            static int FlattenOutValNet = BoardPoints * convEndFiltersValNet;
            static int FlattenOutPolNet = BoardPoints * convEndFiltersPolNet;

            static float DropOut = 0.1f;

            static int inputChannels = 3;

            static int LayerValNeurons = 256; // 256 by default
            static int OutputValValues = 3;

            static int LayerPolNeurons = 256; // 256 by default
            static int OutputPolValues = BoardPoints;

            private readonly nn.Module<Tensor, Tensor> LinLayersValNet;
            private readonly nn.Module<Tensor, Tensor> LinLayersPolNet;


            private nn.Module<Tensor, Tensor> conv_start = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: inputChannels, out_channels: cnnFiltersResBlock, kernel_size: 3,
                padding: 1, padding_mode: PaddingModes.Zeros, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endValNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersValNet, kernel_size: 1, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endPolNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersPolNet, kernel_size: 1, stride: 1);


            private nn.Module<Tensor, Tensor> resNetBlock = new ResNetUnit();

            private nn.Module<Tensor, Tensor> flat_01 = nn.Flatten();

            private nn.Module<Tensor, Tensor> linVal_1 = nn.Linear(FlattenOutValNet, LayerValNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linVal_2 = nn.Linear(LayerValNeurons, OutputValValues, dtype: ScalarType.Float32);

            private nn.Module<Tensor, Tensor> linPol_1 = nn.Linear(FlattenOutPolNet, LayerPolNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linPol_2 = nn.Linear(LayerPolNeurons, OutputPolValues, dtype: ScalarType.Float32);
        }


        public class ComplexNet04 : nn.Module<Tensor, (Tensor, Tensor)> // Based on ComplexNet10
        {
            public ComplexNet04()
            : base(nameof(ComplexNet04))
            {
                List<(string, nn.Module<Tensor, Tensor>)> modulesValNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesValNet.Add(("Flat", nn.Flatten()));
                modulesValNet.Add(("LinVal01", linVal_1));
                modulesValNet.Add(("Relu01", nn.ReLU()));
                modulesValNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesValNet.Add(("LinVal02", linVal_2));
                modulesValNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersValNet = nn.Sequential(modulesValNet);

                List<(string, nn.Module<Tensor, Tensor>)> modulesPolNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesPolNet.Add(("Flat", nn.Flatten()));
                modulesPolNet.Add(("LinPol01", linPol_1));
                modulesPolNet.Add(("Relu01", nn.ReLU()));
                modulesPolNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesPolNet.Add(("LinPol02", linPol_2));
                modulesPolNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersPolNet = nn.Sequential(modulesPolNet);

                RegisterComponents();
            }

            // Alpha - Zero style NET here...
            public override (Tensor, Tensor) forward(Tensor input)
            {
                // Start common block
                using var cnnStart = conv_start.forward(input);
                using var relStart = nn.functional.relu(cnnStart);

                using var outRes01 = resNetBlock.forward(relStart);
                using var outRes02 = resNetBlock.forward(outRes01);
                using var outRes03 = resNetBlock.forward(outRes02);
                using var outRes04 = resNetBlock.forward(outRes03);
                using var outRes05 = resNetBlock.forward(outRes04);

                // Value Net block

                using var outConvEndValNet = conv_endValNet.forward(outRes05);

                // Policy Net block

                using var outConvEndPolNet = conv_endPolNet.forward(outRes05);

                return (LinLayersValNet.forward(outConvEndValNet), LinLayersPolNet.forward(outConvEndPolNet));
            }


            // **************  ResMet Unit.
            class ResNetUnit : nn.Module<Tensor, Tensor>
            {
                public ResNetUnit()
                    : base(nameof(ResNetUnit))
                {
                    RegisterComponents();
                }

                public override Tensor forward(Tensor input)
                {
                    using var resnet = input;

                    using var outputConv01 = ConvModule.forward(input);
                    using var outputRelu01 = nn.functional.relu(outputConv01);

                    using var outputConv02 = ConvModule.forward(outputRelu01);

                    return (nn.functional.relu(outputConv02) + resnet);
                }

                private TorchSharp.Modules.Conv2d ConvModule = nn.Conv2d(dtype: ScalarType.Float32,
                    in_channels: cnnFiltersResBlock, out_channels: cnnFiltersResBlock, kernel_size: 3, padding: 1,
                    padding_mode: PaddingModes.Zeros, stride: 1);
            }

            // Board size
            static int BoardSize = 10;
            static int BoardPoints = BoardSize * BoardSize;

            // Res net params.
            static int cnnFiltersResBlock = 256;

            // End of convolutiobal block.
            static int convEndFiltersValNet = 2; // Reduce end conv filters to 2 (original has 3)
            static int convEndFiltersPolNet = 3; 

            // Variables value block
            static int FlattenOutValNet = BoardPoints * convEndFiltersValNet;
            static int FlattenOutPolNet = BoardPoints * convEndFiltersPolNet;

            static float DropOut = 0.1f;

            static int inputChannels = 3;

            static int LayerValNeurons = 256; // 256 by default
            static int OutputValValues = 3;

            static int LayerPolNeurons = 256; // 256 by default
            static int OutputPolValues = BoardPoints;

            private readonly nn.Module<Tensor, Tensor> LinLayersValNet;
            private readonly nn.Module<Tensor, Tensor> LinLayersPolNet;


            private nn.Module<Tensor, Tensor> conv_start = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: inputChannels, out_channels: cnnFiltersResBlock, kernel_size: 3,
                padding: 1, padding_mode: PaddingModes.Zeros, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endValNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersValNet, kernel_size: 1, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endPolNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersPolNet, kernel_size: 1, stride: 1);


            private nn.Module<Tensor, Tensor> resNetBlock = new ResNetUnit();

            private nn.Module<Tensor, Tensor> flat_01 = nn.Flatten();

            private nn.Module<Tensor, Tensor> linVal_1 = nn.Linear(FlattenOutValNet, LayerValNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linVal_2 = nn.Linear(LayerValNeurons, OutputValValues, dtype: ScalarType.Float32);

            private nn.Module<Tensor, Tensor> linPol_1 = nn.Linear(FlattenOutPolNet, LayerPolNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linPol_2 = nn.Linear(LayerPolNeurons, OutputPolValues, dtype: ScalarType.Float32);
        }

        public class ComplexNet05 : nn.Module<Tensor, (Tensor, Tensor)> // Based on ComplexNet10
        {
            public ComplexNet05()
            : base(nameof(ComplexNet05))
            {
                List<(string, nn.Module<Tensor, Tensor>)> modulesValNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesValNet.Add(("Flat", nn.Flatten()));
                modulesValNet.Add(("LinVal01", linVal_1));
                modulesValNet.Add(("Relu01", nn.ReLU()));
                modulesValNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesValNet.Add(("LinVal02", linVal_2));
                modulesValNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersValNet = nn.Sequential(modulesValNet);

                List<(string, nn.Module<Tensor, Tensor>)> modulesPolNet = new List<(string, nn.Module<Tensor, Tensor>)>();

                modulesPolNet.Add(("Flat", nn.Flatten()));
                modulesPolNet.Add(("LinPol01", linPol_1));
                modulesPolNet.Add(("Relu01", nn.ReLU()));
                modulesPolNet.Add(("Drop", nn.Dropout(DropOut)));
                modulesPolNet.Add(("LinPol02", linPol_2));
                modulesPolNet.Add(("SoftMax", nn.Softmax(1)));

                LinLayersPolNet = nn.Sequential(modulesPolNet);

                RegisterComponents();
            }

            // Alpha - Zero style NET here...
            public override (Tensor, Tensor) forward(Tensor input)
            {
                // Start common block
                using var cnnStart = conv_start.forward(input);
                using var relStart = nn.functional.relu(cnnStart);

                using var outRes01 = resNetBlock.forward(relStart);
                using var outRes02 = resNetBlock.forward(outRes01);
                using var outRes03 = resNetBlock.forward(outRes02);
                using var outRes04 = resNetBlock.forward(outRes03);
                using var outRes05 = resNetBlock.forward(outRes04);

                // Value Net block

                using var outConvEndValNet = conv_endValNet.forward(outRes05);

                // Policy Net block

                using var outConvEndPolNet = conv_endPolNet.forward(outRes05);

                return (LinLayersValNet.forward(outConvEndValNet), LinLayersPolNet.forward(outConvEndPolNet));
            }


            // **************  ResMet Unit.
            class ResNetUnit : nn.Module<Tensor, Tensor>
            {
                public ResNetUnit()
                    : base(nameof(ResNetUnit))
                {
                    RegisterComponents();
                }

                public override Tensor forward(Tensor input)
                {
                    using var resnet = input;

                    using var outputConv01 = ConvModule.forward(input);
                    using var outputRelu01 = nn.functional.relu(outputConv01);

                    using var outputConv02 = ConvModule.forward(outputRelu01);

                    return (nn.functional.relu(outputConv02) + resnet);
                }

                private TorchSharp.Modules.Conv2d ConvModule = nn.Conv2d(dtype: ScalarType.Float32,
                    in_channels: cnnFiltersResBlock, out_channels: cnnFiltersResBlock, kernel_size: 3, padding: 1,
                    padding_mode: PaddingModes.Zeros, stride: 1);
            }

            // Board size
            static int BoardSize = 10;
            static int BoardPoints = BoardSize * BoardSize;

            // Res net params.
            static int cnnFiltersResBlock = 256;

            // End of convolutiobal block.
            static int convEndFiltersValNet = 3; 
            static int convEndFiltersPolNet = 3;

            // Variables value block
            static int FlattenOutValNet = BoardPoints * convEndFiltersValNet;
            static int FlattenOutPolNet = BoardPoints * convEndFiltersPolNet;

            static float DropOut = 0.1f;

            static int inputChannels = 3;

            static int LayerValNeurons = 256; // 256 by default
            static int OutputValValues = 3;

            static int LayerPolNeurons = 256; // 256 by default
            static int OutputPolValues = BoardPoints;

            private readonly nn.Module<Tensor, Tensor> LinLayersValNet;
            private readonly nn.Module<Tensor, Tensor> LinLayersPolNet;


            private nn.Module<Tensor, Tensor> conv_start = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: inputChannels, out_channels: cnnFiltersResBlock, kernel_size: 3,
                padding: 1, padding_mode: PaddingModes.Zeros, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endValNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersValNet, kernel_size: 1, stride: 1);

            private nn.Module<Tensor, Tensor> conv_endPolNet = nn.Conv2d(dtype: ScalarType.Float32,
                in_channels: cnnFiltersResBlock, out_channels: convEndFiltersPolNet, kernel_size: 1, stride: 1);


            private nn.Module<Tensor, Tensor> resNetBlock = new ResNetUnit();

            private nn.Module<Tensor, Tensor> flat_01 = nn.Flatten();

            private nn.Module<Tensor, Tensor> linVal_1 = nn.Linear(FlattenOutValNet, LayerValNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linVal_2 = nn.Linear(LayerValNeurons, OutputValValues, dtype: ScalarType.Float32);

            private nn.Module<Tensor, Tensor> linPol_1 = nn.Linear(FlattenOutPolNet, LayerPolNeurons, dtype: ScalarType.Float32);
            private nn.Module<Tensor, Tensor> linPol_2 = nn.Linear(LayerPolNeurons, OutputPolValues, dtype: ScalarType.Float32);
        }


        public static void CreateZeroUnitedNet(string ZeroNNFileName, string savedDirFilePath, NetVersion netVersion)
        {
            nn.Module<Tensor, (Tensor, Tensor)> zeroNet = GetNetStructure(netVersion: netVersion);
            string savedNetFilePath = $"{savedDirFilePath}{ZeroNNFileName}";
            zeroNet.save(savedNetFilePath);

            //TorchSharp.Modules.Sequential seq = StaticNN.GetCnnNetStructire();
            //string savedNetFilePath = $"{savedDirFilePath}{ZeroNNFileName}";
            //seq.save(savedNetFilePath);
        }

        internal static NetVersion GetUnitedNetEnumFromString(string netVersion)
        {
            foreach (NetVersion version in Enum.GetValues(typeof(NetVersion)))
            {
                if (version.ToString() == netVersion)
                    return version;
            }

            // if string is not found, return first element of enum pNetVersion (index 0)
            return (NetVersion)Enum.GetValues(typeof(NetVersion)).GetValue(0);
        }

        internal static float[,] Create2DArrayFromTensor(Tensor Tensor2D)
        {
            Int64[] tShape = Tensor2D.shape;
            Int64 tensorObjects = tShape[0];
            Int64 objectWidth = tShape[1];

            float[,] arr2D = new float[tensorObjects, objectWidth];

            for (int i = 0; i < tensorObjects; i++)
            {
                for (int j = 0; j < objectWidth; j++)
                {
                    arr2D[i, j] = Tensor2D[i][j].item<float>();
                }
            }

            return arr2D;
        }

    }
}
