﻿using Avalonia.Controls;
using Avalonia.Controls.Shapes;
using Avalonia.Media;
using AvalonRFZero.Libs;
using SkiaSharp;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AvalonRFZero.PRG
{
    internal class GBoard
    {
        readonly Canvas canvas;
        readonly Ellipse MoveTurn;

        byte[] BoardPosition;
        // int[] LastMove = null;

        int[,] protectedSquares = new int[,] { { 0, 0 }, { 0, 1 }, {0, 8}, {0, 9}, { 1, 0}, { 1, 9 },
                { 8, 0 }, { 8, 9 }, { 9, 0 }, { 9, 1 }, {9, 8}, {9, 9}};

        double margine = 50;
        double boardEdgeMargine = 20;
        int boardColumns = 10;

        // Brush brush = new SolidColorBrush(Color.FromRgb(0, 200, 0));
        Brush brush = new SolidColorBrush(Colors.LightGreen);

        Brush brushEdge = new SolidColorBrush(Color.FromRgb(255, 0, 0));
        Brush protectedSquareBrush = new SolidColorBrush(Color.FromRgb(100, 150, 0));
        Brush whiteFigBrush = new SolidColorBrush(Color.FromRgb(255, 255, 255));
        Brush blackFigBrush = new SolidColorBrush(Color.FromRgb(0, 0, 0));
        Brush lastMoveBrush = new SolidColorBrush(Color.FromRgb(0, 0, 255));

        public bool AnimationIsRunning = false;

        int[] LastMove = null;

        int animPhasesTotal = 20;
        Progress<AnimInfo> progressAnimation;


        public GBoard(Canvas canvas, Ellipse moveTurm) 
        {
            this.canvas = canvas;
            this.MoveTurn = moveTurm;

            DrawBoard();

            progressAnimation = new Progress<AnimInfo>();
            progressAnimation.ProgressChanged += ProgressAnimation_ProgressChanged;

            this.MoveTurn.IsVisible = false;
        }

        private void ProgressAnimation_ProgressChanged(object sender, AnimInfo e)
        {
            DrawBoarAnimationFrame(this.canvas, animFigureList: e.AnimFigureList, 
                lastMove: e.lastMove, NewPosition: e.NewPosition, animPhase: e.AnimPhase);
        }

        void AnimationLoop(IProgress<AnimInfo> progressAnimation, int[] lastMove, byte[] NewPosition, byte[] OldPosition)
        {
            if (lastMove == null || OldPosition == null) 
            {
                progressAnimation.Report(new AnimInfo(AnimPhase: 0, AnimFigureList: null, 
                    lastMove: null, NewPosition: NewPosition));
                
                return;
            }

            List<int[]> animFigureList = FindFiguresToAnimate(NewPosition, OldPosition);

            for (int animPhase = 0; animPhase < this.animPhasesTotal; animPhase++)
            {
                progressAnimation.Report(new AnimInfo(animPhase, AnimFigureList: animFigureList, 
                    lastMove: lastMove, NewPosition: NewPosition));
                
                Task.Delay(20).Wait();
            }
        }

        public void DrawPosition(byte[] Position)
        {
            BoardPosition = Position;

            DrawMainBoard(canvas);
            DrawBoardFigures(canvas);

            DrawMoveTurn(BoardPosition);
        }

        public async void DrawPositionWithAnimation(byte[] Position, int[] lastMove) 
        {
            AnimationIsRunning = true;
            this.LastMove = lastMove;

            await Task.Run(() => AnimationLoop(progressAnimation, lastMove: lastMove, NewPosition: Position, OldPosition: BoardPosition));

            if (lastMove != null)
                DrawLastMoveMark(canvas, row: lastMove[0], col: lastMove[1]);

            // We need to set BoardPosition after animation is done.
            BoardPosition = Position;

            AnimationIsRunning = false;

            DrawMoveTurn(BoardPosition);
        }

        internal void DrawBoard()
        {
            DrawMainBoard(canvas);
            DrawBoardFigures(canvas);

            if (this.LastMove != null)
                DrawLastMoveMark(canvas, row: LastMove[0], col: LastMove[1]);

            // DrawMoveTurn(BoardPosition);
        }

        internal void DrawMoveTurn(byte[] Position) 
        {
            if (Position == null)
                return;

            if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.WhiteTurn) 
            {
                this.MoveTurn.IsVisible = true;
                this.MoveTurn.Fill = new SolidColorBrush(Colors.White);
            }

            else if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.BlackTurn) 
            {
                this.MoveTurn.IsVisible = true;
                this.MoveTurn.Fill = new SolidColorBrush(Colors.Black);
            }
            
            else
                this.MoveTurn.IsVisible = false;
        }

        void DrawMainBoard(Canvas canvas) 
        {
            ClearBoard();

            double spaceX = canvas.Bounds.Width;
            double spaceY = canvas.Bounds.Height;

            double boardLenth = Math.Min(spaceX, spaceY) - 2 * margine;

            double posX = (spaceX - boardLenth) / 2;
            double posY = (spaceY - boardLenth) / 2;

            double boardEdgeLength = boardLenth + 2 * boardEdgeMargine;
            double posXEdge = posX - boardEdgeMargine;
            double posYEdge = posY - boardEdgeMargine;

            Rectangle recBoardWirhEdge = new Rectangle
            {
                Tag = "Big",
                Height = boardEdgeLength,
                Width = boardEdgeLength,
                Stroke = Brushes.Black,
                StrokeThickness = 1,
                Fill = brushEdge
            };

            Canvas.SetLeft(recBoardWirhEdge, posXEdge);
            Canvas.SetTop(recBoardWirhEdge, posYEdge);

            canvas.Children.Add(recBoardWirhEdge);


            Rectangle recBoard = new Rectangle
            {
                Tag = "Main",
                Height = boardLenth,
                Width = boardLenth,
                Stroke = Brushes.Black,
                StrokeThickness = 1,
                Fill = brush
            };

            Canvas.SetLeft(recBoard, posX);
            Canvas.SetTop(recBoard, posY);

            canvas.Children.Add(recBoard);

            // Draw protected sqiares
            for (int i = 0; i < protectedSquares.GetLength(0); i++)
            {
                DrawProtectedSquares(canvas, protectedSquares[i, 0], protectedSquares[i, 1],
                    protectedSquareBrush, boardLenth, boardColumns, posX, posY);
            }

            // Draw vertical and horizontal lines here...
            double squareWidth = boardLenth / boardColumns;

            for (int i = 1; i < boardColumns; i++)
            {
                Line lineVert = new Line
                {
                    StartPoint = new Avalonia.Point(posX + squareWidth * i, posY),
                    EndPoint = new Avalonia.Point(posX + squareWidth * i, posY + boardLenth),
                    Stroke = Brushes.Red,
                    StrokeThickness = 3
                };

                canvas.Children.Add(lineVert);
            }

            for (int i = 1; i < boardColumns; i++)
            {
                Line lineVert = new Line
                {
                    StartPoint = new Avalonia.Point(posX, posY + squareWidth * i),
                    EndPoint = new Avalonia.Point(posX + boardLenth, posY + squareWidth * i),
                    Stroke = Brushes.Red,
                    StrokeThickness = 3
                };

                canvas.Children.Add(lineVert);
            }
        }

        void DrawBoardFigures(Canvas canvas) 
        {
            // Draw white and black figires om the board

            double spaceX = canvas.Bounds.Width;
            double spaceY = canvas.Bounds.Height;

            double boardLenth = Math.Min(spaceX, spaceY) - 2 * margine;

            double posX = (spaceX - boardLenth) / 2;
            double posY = (spaceY - boardLenth) / 2;


            if (BoardPosition != null)
            {
                for (int row = 0; row < boardColumns; row++)
                {
                    for (int col = 0; col < boardColumns; col++)
                    {
                        if (StaticMath8Stars.GetPosSquare(BoardPosition, row, col) == StaticMath8Stars.WhiteFig)
                            DrawSimpleUnitOnGraphicsObject(canvas, row: row, col: col, boardSizePix: boardLenth, currBrush: whiteFigBrush, boardColumns: boardColumns,
                                StartX: posX, StartY: posY);
                        if (StaticMath8Stars.GetPosSquare(BoardPosition, row, col) == StaticMath8Stars.BlackFig)
                            DrawSimpleUnitOnGraphicsObject(canvas, row: row, col: col, boardSizePix: boardLenth, currBrush: blackFigBrush, boardColumns: boardColumns,
                                StartX: posX, StartY: posY);
                    }
                }
            }
        }

        void DrawBoarAnimationFrame(Canvas canvas, int[] lastMove, byte[] NewPosition, List<int[]> animFigureList, int animPhase) 
        {
            if (lastMove == null) 
            {
                BoardPosition = NewPosition;

                DrawBoard();
                return;
            }

            int LastMoveColor = StaticMath8Stars.GetPosSquare(NewPosition, Row: lastMove[0], Col: lastMove[1]);

            (Brush BrushStart, Brush BrushEnd) = GetStartEndAnimationBrush(BrushWhite: whiteFigBrush, BrushBlack: blackFigBrush,
                LastMoveColor: LastMoveColor);

                // We need to redraw Main board every frame of animation.
                DrawMainBoard(canvas);

                (double boardLenth, double startX, double startY) = CalculateMainBoardParams(canvas);

                for (int row = 0; row < StaticMath8Stars.BoardSize; row++)
                {
                    for (int col = 0; col < StaticMath8Stars.BoardSize; col++)
                    {
                        if (FigureNeedsToBeAnimated(row, col, animFigureList) == false)
                        {
                            if (StaticMath8Stars.GetPosSquare(NewPosition, row, col) == StaticMath8Stars.WhiteFig)
                                DrawSimpleUnitOnGraphicsObject(canvas, row, col, whiteFigBrush, boardColumns: StaticMath8Stars.BoardSize,
                                    boardSizePix: boardLenth, StartX: startX, StartY: startY);
                            if (StaticMath8Stars.GetPosSquare(NewPosition, row, col) == StaticMath8Stars.BlackFig)
                                DrawSimpleUnitOnGraphicsObject(canvas, row, col, blackFigBrush, boardColumns: StaticMath8Stars.BoardSize,
                                    boardSizePix: boardLenth, StartX: startX, StartY: startY);
                        }

                        else
                        {
                            DrawSimpleUnitAnimationPhase(canvas, row, col, BrushStart: BrushStart, BrushEnd: BrushEnd,
                                boardSizePix: boardLenth, animPhase: animPhase, animPhaseTotal: this.animPhasesTotal, StartX: startX, StartY: startY);
                        }
                    }
                }
        }

        (double boardLenth, double startX, double startY) CalculateMainBoardParams(Canvas canvas) 
        {
            double spaceX = canvas.Bounds.Width;
            double spaceY = canvas.Bounds.Height;

            double boardLenth = Math.Min(spaceX, spaceY) - 2 * margine;

            double startX = (spaceX - boardLenth) / 2;
            double startY = (spaceY - boardLenth) / 2;

            return (boardLenth, startX, startY);
        }

        List<int[]> FindFiguresToAnimate(byte[] NewPosition, byte[] OldPosition)
        {
            List<int[]> animFigureList = new List<int[]>();

            for (int row = 0; row < StaticMath8Stars.BoardSize; row++)
            {
                for (int col = 0; col < StaticMath8Stars.BoardSize; col++)
                {
                    int NewFig = StaticMath8Stars.GetPosSquare(Position: NewPosition, Row: row, Col: col);
                    int OldFig = StaticMath8Stars.GetPosSquare(Position: OldPosition, Row: row, Col: col);

                    if (NewFig != OldFig && OldFig != StaticMath8Stars.EmptyFig)
                    {
                        animFigureList.Add(new int[] { row, col });
                    }
                }
            }

            return animFigureList;
        }

        bool FigureNeedsToBeAnimated(int row, int col, List<int[]> animFigureList)
        {
            for (int i = 0; i < animFigureList.Count; i++)
            {
                if (animFigureList[i][0] == row && animFigureList[i][1] == col)
                    return true;
            }

            return false;
        }

        (Brush BrushStart, Brush BrushEnd) GetStartEndAnimationBrush(Brush BrushWhite, Brush BrushBlack, int LastMoveColor)
        {
            Brush BrushStart;
            Brush BrushEnd;

            if (LastMoveColor == StaticMath8Stars.WhiteFig)
            {
                BrushStart = BrushBlack;
                BrushEnd = BrushWhite;
            }

            else
            {
                BrushStart = BrushWhite;
                BrushEnd = BrushBlack;
            }

            return (BrushStart: BrushStart, BrushEnd: BrushEnd);
        }

        void DrawProtectedSquares(Canvas canvas, int row, int col, Brush currBrush, double boardLenth, int boardColumns,
         double StartX, double StartY)
        {
            double SquareSize = boardLenth / boardColumns;

            double xCenter = StartX + (col * SquareSize) + SquareSize / 2;
            double yCenter = StartY + boardLenth - ((row + 1) * SquareSize) + SquareSize / 2;

            double X = xCenter - (SquareSize / 2);
            double Y = yCenter - (SquareSize / 2);

            Rectangle recProtSquare = new Rectangle
            {
                Tag = "ProtectedSquare",
                Height = SquareSize,
                Width = SquareSize,
                Stroke = Brushes.Black,
                StrokeThickness = 1,
                Fill = currBrush
            };

            Canvas.SetLeft(recProtSquare, X);
            Canvas.SetTop(recProtSquare, Y);

            canvas.Children.Add(recProtSquare);
        }

        void DrawSimpleUnitOnGraphicsObject(Canvas canvas, int row, int col, Brush currBrush, int boardColumns,
            double boardSizePix, double StartX, double StartY)
        {
            double scaleY = 0.9f;
            double scaleX = 0.9f;
            double SquareSize = boardSizePix / boardColumns;

            double figureBoxWidth = scaleX * boardSizePix / boardColumns;
            double figureBoxHeight = scaleY * boardSizePix / boardColumns;

            double xCenter = StartX + (col * SquareSize) + SquareSize / 2;
            double yCenter = StartY + boardSizePix - ((row + 1) * SquareSize) + SquareSize / 2;

            double X = xCenter - (figureBoxWidth / 2);
            double Y = yCenter - (figureBoxHeight / 2);

            Ellipse circle = new Ellipse
            {
                Tag = "Circle",
                Height = figureBoxWidth,
                Width = figureBoxHeight,
                Fill = currBrush
            };

            Canvas.SetLeft(circle, X);
            Canvas.SetTop(circle, Y);

            canvas.Children.Add(circle);
        }

        void DrawSimpleUnitAnimationPhase(Canvas canvas, int row, int col,
             Brush BrushStart, Brush BrushEnd, int animPhase, int animPhaseTotal, double boardSizePix, double StartX, double StartY)
        {
            // animPhase can be like 0,1,2,3,4,5,6,7,8,9: 10 animation Phases total 
            // we need to calculate scaleX and BrushColor for any animation phase.

            double scaleX_Max = 0.9f;
            double scaleY_Max = 0.9f;

            double alphaStep = 3.14f / (animPhaseTotal - 1);
            double cosValue = Math.Cos(alphaStep * animPhase);

            double scaleX = scaleX_Max * cosValue;
            double scaleY = scaleY_Max;
            double SquareSize = boardSizePix / StaticMath8Stars.BoardSize;

            Brush currBrush = BrushStart;

            if (cosValue > 0)
                currBrush = BrushStart;
            else
                currBrush = BrushEnd;

            double figureBoxWidth = Math.Abs(scaleX * boardSizePix / StaticMath8Stars.BoardSize);
            double figureBoxHeight = scaleY * boardSizePix / StaticMath8Stars.BoardSize;

            double xCenter = StartX + (col * SquareSize) + SquareSize / 2;
            double yCenter = StartY + boardSizePix - ((row + 1) * SquareSize) + SquareSize / 2;

            double X = xCenter - (figureBoxWidth / 2);
            double Y = yCenter - (figureBoxHeight / 2);

            // g.FillEllipse(currBrush, X, Y, figureBoxWidth, figureBoxHeight);

            Ellipse circle = new Ellipse
            {
                Tag = "Circle",
                Height = figureBoxHeight,
                Width = figureBoxWidth,
                Fill = currBrush
            };

            Canvas.SetLeft(circle, X);
            Canvas.SetTop(circle, Y);

            canvas.Children.Add(circle);
        }

        void DrawLastMoveMark(Canvas canvas, int row, int col)
        {
            double spaceX = canvas.Bounds.Width;
            double spaceY = canvas.Bounds.Height;

            double boardLenth = Math.Min(spaceX, spaceY) - 2 * margine;

            double StartX = (spaceX - boardLenth) / 2;
            double StartY = (spaceY - boardLenth) / 2;

            double scaleY = 0.2;
            double scaleX = 0.2;
            double SquareSize = boardLenth / boardColumns;

            double figureBoxWidth = scaleX * boardLenth / boardColumns;
            double figureBoxHeight = scaleY * boardLenth / boardColumns;

            double xCenter = StartX + (col * SquareSize) + SquareSize / 2;
            double yCenter = StartY + boardLenth - ((row + 1) * SquareSize) + SquareSize / 2;

            double X = xCenter - (figureBoxWidth / 2);
            double Y = yCenter - (figureBoxHeight / 2);

            Ellipse circle = new Ellipse
            {
                Tag = "Circle",
                Height = figureBoxWidth,
                Width = figureBoxHeight,
                Fill = lastMoveBrush
            };

            Canvas.SetLeft(circle, X);
            Canvas.SetTop(circle, Y);

            canvas.Children.Add(circle);
        }



        internal void ClearBoard()
        {
            canvas.Children.Clear();
        }

        public (int row, int col) GetRowColFromMouseCoordinates(double mouseX, double mouseY)
        {

            double boardWidth = Math.Min(canvas.Bounds.Width - margine, canvas.Bounds.Height - margine);

            // int boardWidth = Math.Min(Canvas.Width, Canvas.Height);

            double xCenter = canvas.Bounds.Width / 2;
            double yCenter = canvas.Bounds.Height / 2;

            double visualSquareSize = boardWidth / boardColumns;

            //float xClickMin = (Canvas.Width - boardWidth) / 2;
            //float xClickMax = xClickMin + boardWidth;

            double xClickMin = (xCenter - boardWidth / 2);
            double xClickMax = (xCenter + boardWidth / 2);

            // float yClickMin = (Canvas.Height - boardWidth) / 2;
            // float yClickMax = yClickMin + boardWidth;

            double yClickMin = (yCenter - boardWidth / 2);
            double yClickMax = (yCenter + boardWidth / 2);

            if (mouseX < xClickMin || mouseX > xClickMax || mouseY < yClickMin || mouseY > yClickMax)
                return (-1, -1); // Point is out of the board

            int col = (int)Math.Truncate((mouseX - xClickMin) / visualSquareSize);
            int row = (boardColumns - 1) - (int)Math.Truncate((mouseY - yClickMin) / visualSquareSize);

            return (row, col);
        }

        class AnimInfo
        {
            public int AnimPhase { set; get; }
            public List<int[]> AnimFigureList { set; get; }
            public int[] lastMove {  set; get; }
            public byte[] NewPosition { set; get; }

            public AnimInfo(int AnimPhase, List<int[]> AnimFigureList, int[] lastMove, byte[] NewPosition)
            {
                this.AnimPhase = AnimPhase;
                this.AnimFigureList = AnimFigureList;
                this.lastMove = lastMove;
                this.NewPosition = NewPosition;
            }
        }
    }
}
