﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Avalonia.Controls;
using AvalonRFZero.Libs;

namespace AvalonRFZero.PRG
{
    internal class HumanComputerMatch
    {
        const string HumanComputerSetupFile = @"SETUP\HumanEngineSetup.txt";

        IEngine EngineWhite = null;
        IEngine EngineBlack = null;

        GBoard gBoard;
        // DataGridView dataGridViewMovesList;
        Label HumanCompMessage;


        BoardMoves movesVisualiser;

        byte[] Position;
        int[] HumanMove = new int[2];

        public HumanComputerMatch(GBoard gBoard, BoardMoves movesVisualiser, Label HumanCompMessage,
            int WhitePlayerIndex, int BlackPlayerIndex)
        {
            this.gBoard = gBoard;
            this.HumanCompMessage = HumanCompMessage;

            HumanMove[0] = -1;
            HumanMove[1] = -1;

            this.movesVisualiser = movesVisualiser;

            (EngineWhite, EngineBlack) = SetEngineByIndex(WhitePlayerIndex, BlackPlayerIndex);

            Progress<PosReport> progressPosition = new Progress<PosReport>();
            progressPosition.ProgressChanged += ProgressPosition_ProgressChanged;

            Progress<ResultReport> progressMessage = new Progress<ResultReport>();
            progressMessage.ProgressChanged += MessageReport_ProgressChanged;

            // movesVisualiser.ClearTable();

            Task.Run(() => TakeSingleGame(progressPosition, progressMessage));
        }

        private void MessageReport_ProgressChanged(object sender, ResultReport e)
        {
            this.HumanCompMessage.IsVisible = true;
            this.HumanCompMessage.Content = e.Message;
        }

        private void ProgressPosition_ProgressChanged(object sender, PosReport e)
        {
            gBoard.DrawPositionWithAnimation(e.PositionCurrent, e.LastMove);
            movesVisualiser.AddMoveToTable(move: e.LastMove, moveturn: e.LastMoveTurn);

        }

        public async void TakeSingleGame(IProgress<PosReport> progressPosition, IProgress<ResultReport> progressResult)
        {
            // movesVisualiser.ClearTable();
            Position = StaticMath8Stars.CreateStandartInitialPosition();
            progressPosition.Report(new PosReport(Position: Position, LastMoveMade: null, LastMoveTurn: -1));

            Task.Delay(1000).Wait(); // Wait 1000 milseconds before next. 

            while (Position[StaticMath8Stars.MoveTurnIndex] != StaticMath8Stars.GameOver)
            {
                if (NowIsHumanTurn() == true) // Make human move here...
                {
                    if (HumanMove[0] != -1 && HumanMove[1] != -1)
                    {
                        int moveTurn = Position[StaticMath8Stars.MoveTurnIndex];
                        StaticMath8Stars.MakeSquareMove(Position, HumanMove[0], HumanMove[1], Position[StaticMath8Stars.MoveTurnIndex]);

                        gBoard.AnimationIsRunning = true;
                        progressPosition.Report(new PosReport(Position, HumanMove, moveTurn));

                        HumanMove[0] = -1;
                        HumanMove[1] = -1;
                    }
                }
                else // Computer turn is here... 
                {
                    int[] bestMove;

                    if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.WhiteFig)
                    {
                        bestMove = EngineWhite.ShowBestMove(Position);
                    }
                    else
                    {
                        bestMove = EngineBlack.ShowBestMove(Position);
                    }

                    // We are making board move here....
                    int moveTurn = Position[StaticMath8Stars.MoveTurnIndex];

                    // We make computer move here... 
                    StaticMath8Stars.MakeSquareMove(Position, bestMove[0], bestMove[1], Position[StaticMath8Stars.MoveTurnIndex]);
                    gBoard.AnimationIsRunning = true;
                    progressPosition.Report(new PosReport(Position, bestMove, moveTurn));
                }


                // **** This is for debug only.
                bool anim = gBoard.AnimationIsRunning;

                Stopwatch stopwatch = Stopwatch.StartNew();
                // Need to wait till the board animation is finished. 
                await Task.Run(() => WaitForBoardAnimation(gBoard));
                stopwatch.Stop();

                double ms = stopwatch.ElapsedMilliseconds;


                Task.Delay(50).Wait(); // Wait 50 more milseconds before next. 
            }

            progressResult.Report(new ResultReport(Position));
        }

        void WaitForBoardAnimation(GBoard gBoard)
        {
            while (gBoard.AnimationIsRunning == true)
            {
                Task.Delay(25).Wait();
            }
        }

        public static void CreateHumanEngineList(ComboBox playerWhite, ComboBox playerBlack)
        {
            playerWhite.Items.Clear();
            playerBlack.Items.Clear();

            XmlDocument doc = new XmlDocument();
            doc.Load(HumanComputerSetupFile);

            XmlNode nodeEngineRoot = doc.DocumentElement.SelectSingleNode("/Engines");

            foreach (XmlNode childNode in nodeEngineRoot.ChildNodes)
            {
                playerWhite.Items.Add(childNode.Attributes["EngineName"].InnerText);
                playerBlack.Items.Add(childNode.Attributes["EngineName"].InnerText);
            }

            playerWhite.SelectedIndex = 0;
            playerBlack.SelectedIndex = 1;
        }

        (IEngine EngineWhite, IEngine EngineBlack) SetEngineByIndex(int WhitePlayerIndex, int BlackPlayerIndex)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(HumanComputerSetupFile);

            XmlNode nodeEngineRoot = doc.DocumentElement.SelectSingleNode("/Engines");

            int currIndex = 0;
            foreach (XmlNode childNode in nodeEngineRoot.ChildNodes)
            {
                if (currIndex == WhitePlayerIndex)
                {
                    EngineWhite = EngineSerial.CreateEngineFromXmlNode(nodeEngine: childNode);
                }

                if (currIndex == BlackPlayerIndex)
                {
                    EngineBlack = EngineSerial.CreateEngineFromXmlNode(nodeEngine: childNode);
                }

                currIndex++;
            }

            return (EngineWhite: EngineWhite, EngineBlack: EngineBlack);
        }

        public void MakeHumanrMove(int row, int col)
        {
            if (row == -1 || col == -1)
                return; // Invalid point here.

            if (gBoard.AnimationIsRunning == true)
                return; // We do not make the move while animation is going.

            if (NowIsHumanTurn() == true && HumanMakeValidMove(row, col))
            {
                HumanMove[0] = row;
                HumanMove[1] = col;
            }
        }

        bool NowIsHumanTurn()
        {
            if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.WhiteFig && EngineWhite == null)
                return true;

            if (Position[StaticMath8Stars.MoveTurnIndex] == StaticMath8Stars.BlackFig && EngineBlack == null)
                return true;
            return false;
        }

        bool HumanMakeValidMove(int row, int col)
        {
            if (StaticMath8Stars.TrySquareMove(Position, row, col, Position[StaticMath8Stars.MoveTurnIndex]) == true)
                return true;
            else
                return false;
        }

        internal class PosReport
        {
            public byte[] PositionCurrent { set; get; } = new byte[StaticMath8Stars.PosArrayLenth];
            public int[] LastMove { set; get; } = new int[2];
            public int LastMoveTurn { set; get; }

            public PosReport(byte[] Position, int[] LastMoveMade, int LastMoveTurn)
            {
                Array.Copy(Position, PositionCurrent, Position.Length);

                if (LastMoveMade != null)
                {
                    Array.Copy(LastMoveMade, LastMove, LastMoveMade.Length);
                    this.LastMoveTurn = LastMoveTurn;
                }
                else
                {
                    LastMove = null;
                    this.LastMoveTurn = -1;
                }
            }
        }

        internal class ResultReport
        {
            public string Message { set; get; }

            public ResultReport(byte[] Position)
            {
                (StaticMath8Stars.gameResult GameResult, int WhiteCount, int BlackCount) = CalcGameResult(Position);

                string gameRes;

                if (GameResult == StaticMath8Stars.gameResult.WhiteWins)
                    gameRes = "White wins";
                else if (GameResult == StaticMath8Stars.gameResult.BlackWins)
                    gameRes = "Black wins";
                else
                    gameRes = "Draw";

                Message = $"{gameRes} {WhiteCount} : " +
                    $"{BlackCount}!";
            }

            static public (StaticMath8Stars.gameResult GameResult, int WhiteCount, int BlackCount) CalcGameResult(byte[] finalPosition)
            {
                (int whiteCount, int blackCount) = StaticMath8Stars.StaticResult(finalPosition);

                if (whiteCount - blackCount > 0) // White wins
                {
                    return (GameResult: StaticMath8Stars.gameResult.WhiteWins, WhiteCount: whiteCount, BlackCount: blackCount);
                }
                else if (whiteCount - blackCount < 0) // Black wins
                {
                    return (GameResult: StaticMath8Stars.gameResult.BlackWins, WhiteCount: whiteCount, BlackCount: blackCount);
                }
                else // Draw.
                {
                    return (GameResult: StaticMath8Stars.gameResult.Draw, WhiteCount: whiteCount, BlackCount: blackCount);
                }
            }
        }

        //public static void CreateHumanEngineList(ComboBox player1, ComboBox player2)
        //{
        //    player1.Items.Clear();
        //    player2.Items.Clear();

        //    XmlDocument doc = new XmlDocument();
        //    doc.Load(HumanComputerSetupFile);

        //    XmlNode nodeEngineRoot = doc.DocumentElement.SelectSingleNode("/Engines");

        //    foreach (XmlNode childNode in nodeEngineRoot.ChildNodes)
        //    {
        //        player1.Items.Add(childNode.Attributes["EngineName"].InnerText);
        //        player2.Items.Add(childNode.Attributes["EngineName"].InnerText);
        //    }

        //    player1.SelectedIndex = 0;
        //    player2.SelectedIndex = 0;
        //}
    }
}
